﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows;
using System.Windows.Controls;

namespace Praktikum_3
{
    class Zelle
    {
        static List<Zelle> zellen = new List<Zelle>();

        TextBox textBox;
        string adresse;
        int ergebnis;
        string eingabe;
        Exception fehler;

        // besser: HashSet
        List<Zelle> wenIchVerwende = new List<Zelle>();
        List<Zelle> werMichVerwendet = new List<Zelle>();

        public Zelle(Grid grid, int zeile, int spalte)
        {
            textBox = new TextBox();
            grid.Children.Add(textBox);
            Grid.SetRow(textBox, zeile + 1);
            Grid.SetColumn(textBox, spalte + 1);

            adresse = new string((char)('A' + spalte), 1) + (zeile + 1);
            textBox.Text = eingabe = "0";

            textBox.GotFocus += HatFocusBekommen;
            textBox.LostFocus += HatFocusVerloren;

            zellen.Add(this);
        }

        void HatFocusBekommen(object sender, RoutedEventArgs e)
        {
            textBox.Text = eingabe;
        }

        void HatFocusVerloren(object sender, RoutedEventArgs e)
        {
            eingabe = textBox.Text;
            Berechne(this, true);
        }

        void Berechne(Zelle auslöser, bool istKopfDesUpdates)
        {
            fehler = null;

            // Verweise können sich nur bei istKopfDesUpdates geändert haben,
            // denn nur dort wurde die Formel editiert.
            if (istKopfDesUpdates)
            {
                // alle bisherigen Verweise vergessen,
                // um sie dann weiter unten aktuell einzusammeln.
                foreach (Zelle zelle in wenIchVerwende)
                {
                    zelle.werMichVerwendet.Remove(this);
                }
                wenIchVerwende.Clear();
            }

            try
            {
                if (!istKopfDesUpdates && auslöser == this)
                {
                    throw new ZirkelbezugException();
                }

                // professionell nicht mit if und Split usw.,
                // sondern mit regulären Ausdrücken (RegEx)
                if (eingabe.StartsWith("="))
                {
                    string[] teile = eingabe.Substring(1).Split('+');
                    System.Diagnostics.Trace.WriteLine("In " + adresse + " " + teile.Length + " Teile gefunden.");

                    ergebnis = 0;
                    foreach (string teil in teile)
                    {
                        // geht besser mit einem Dictionary statt einer List
                        Zelle zelle = zellen.Find(z => z.adresse == teil);

                        if (zelle == null)
                        {
                            throw new FormelfehlerException();
                        }

                        if (zelle.fehler != null)
                        {
                            throw zelle.fehler;
                        }

                        ergebnis += zelle.ergebnis;

                        if (istKopfDesUpdates && !wenIchVerwende.Contains(zelle))
                        {
                            wenIchVerwende.Add(zelle);
                            zelle.werMichVerwendet.Add(this);
                        }
                    }
                }
                else
                {
                    ergebnis = int.Parse(eingabe);
                }
                System.Diagnostics.Trace.WriteLine("Das Ergebnis von " + adresse + " ist " + ergebnis + ".");
            }
            catch (Exception ex)
            {
                fehler = ex;
            }

            if (fehler != null)
            {
                textBox.Text = fehler.Message;
            }
            else
            {
                textBox.Text = ergebnis.ToString();
            }

            if(!(fehler is ZirkelbezugException))
            {
                foreach (Zelle v in werMichVerwendet)
                {
                    v.Berechne(auslöser, false);
                }
            }
        }
    }

    class FormelfehlerException: ApplicationException
    {
        public FormelfehlerException()
            : base("Fehler in Formel")
        { }
    }

    class ZirkelbezugException : ApplicationException
    {
        public ZirkelbezugException()
            : base("Zirkelbezug")
        { }
    }
}
